from typing import TYPE_CHECKING, Generic, TypeVar, cast

from vintasend.constants import NotificationTypes
from vintasend.services.notification_adapters.base import BaseNotificationAdapter
from vintasend.services.notification_backends.base import BaseNotificationBackend
from vintasend.services.notification_template_renderers.base import BaseNotificationTemplateRenderer


if TYPE_CHECKING:
    from vintasend.services.dataclasses import Notification
    from vintasend.services.notification_service import NotificationContextDict


B = TypeVar("B", bound=BaseNotificationBackend)
T = TypeVar("T", bound=BaseNotificationTemplateRenderer)


class FakeInAppAdapter(Generic[B, T], BaseNotificationAdapter[B, T]):
    notification_type = NotificationTypes.IN_APP
    sent_emails: list[tuple["Notification", "NotificationContextDict"]] = []

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.sent_emails = []

    def send(self, notification: "Notification", context: "NotificationContextDict") -> None:
        self.template_renderer.render(notification, context)
        self.sent_emails.append((notification, context))



import asyncio
import datetime
import json
import os
import uuid
from decimal import Decimal

from vintasend.constants import NotificationStatus, NotificationTypes
from vintasend.exceptions import NotificationNotFoundError
from vintasend.services.dataclasses import Notification, UpdateNotificationKwargs
from vintasend.services.notification_backends.asyncio_base import AsyncIOBaseNotificationBackend
from vintasend.services.notification_backends.base import BaseNotificationBackend


class FakeFileBackend(BaseNotificationBackend):
    notifications: list[Notification]
    database_file_name: str

    def __init__(self, database_file_name: str = "notifications.json", **kwargs):
        super().__init__(database_file_name=database_file_name, **kwargs)
        self.database_file_name = database_file_name
        try:
            notifications_file = open(self.database_file_name, encoding="utf-8")
        except FileNotFoundError:
            self.notifications = []
            return
        try:
            self.notifications = [
                self._convert_json_to_notification(n) for n in json.load(notifications_file)
            ]
            notifications_file.close()
        except json.JSONDecodeError:
            self.notifications = []
            return

    def clear(self):
        self.notifications = []
        try:
            os.remove(self.database_file_name)
        except FileNotFoundError:
            pass

    def get_future_notifications(self, page: int, page_size: int) -> list[Notification]:
        return self.__paginate_notifications(self.get_all_future_notifications(), page, page_size)

    def get_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID, page: int, page_size: int
    ) -> list[Notification]:
        return self.__paginate_notifications(
            self.get_all_future_notifications_from_user(user_id), page, page_size
        )

    def get_all_future_notifications(self) -> list[Notification]:
        return [
            n
            for n in self.notifications
            if n.status == NotificationStatus.PENDING_SEND.value
            and (
                n.send_after is not None
                and n.send_after > datetime.datetime.now(tz=datetime.timezone.utc)
            )
        ]

    def get_all_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID
    ) -> list[Notification]:
        return [
            n
            for n in self.notifications
            if n.status == NotificationStatus.PENDING_SEND.value
            and (
                n.send_after is not None
                and n.send_after > datetime.datetime.now(tz=datetime.timezone.utc)
            )
            and str(n.user_id) == str(user_id)
        ]

    def get_all_pending_notifications(self) -> list[Notification]:
        return [
            n
            for n in self.notifications
            if n.status == NotificationStatus.PENDING_SEND.value
            and (
                n.send_after is None
                or n.send_after <= datetime.datetime.now(tz=datetime.timezone.utc)
            )
        ]

    def _convert_notification_to_json(self, notification: Notification) -> dict:
        return {
            "id": str(notification.id),
            "user_id": str(notification.user_id),
            "notification_type": notification.notification_type,
            "title": notification.title,
            "body_template": notification.body_template,
            "context_name": notification.context_name,
            "context_kwargs": notification.context_kwargs,
            "send_after": notification.send_after.isoformat() if notification.send_after else None,
            "subject_template": notification.subject_template,
            "preheader_template": notification.preheader_template,
            "status": notification.status,
            "context_used": notification.context_used,
            "adapter_extra_parameters": notification.adapter_extra_parameters,
        }

    def _convert_json_to_notification(self, notification: dict) -> Notification:
        return Notification(
            id=notification["id"],
            user_id=notification["user_id"],
            notification_type=notification["notification_type"],
            title=notification["title"],
            body_template=notification["body_template"],
            context_name=notification["context_name"],
            context_kwargs=notification["context_kwargs"],
            send_after=(
                datetime.datetime.fromisoformat(notification["send_after"])
                if notification["send_after"]
                else None
            ),
            subject_template=notification["subject_template"],
            preheader_template=notification["preheader_template"],
            status=notification["status"],
            context_used=notification.get("context_used"),
            adapter_extra_parameters=notification.get("adapter_extra_parameters"),
        )

    def _store_notifications(self):
        json_output_file = open(self.database_file_name, "w", encoding="utf-8")
        json.dump(
            [self._convert_notification_to_json(n) for n in self.notifications],
            json_output_file,
        )
        json_output_file.close()

    def get_pending_notifications(self, page: int, page_size: int) -> list[Notification]:
        # page is 1-indexed
        return self.get_all_pending_notifications()[
            ((page - 1) * page_size) : ((page - 1) * page_size) + page_size
        ]

    def persist_notification(
        self,
        user_id: uuid.UUID | str | int,
        notification_type: str,
        title: str,
        body_template: str,
        context_name: str,
        context_kwargs: dict[str, uuid.UUID | str | int],
        send_after: datetime.datetime | None,
        subject_template: str,
        preheader_template: str,
        adapter_extra_parameters: dict | None = None,
    ) -> Notification:
        notification = Notification(
            id=str(uuid.uuid4()),
            user_id=user_id,
            notification_type=notification_type,
            title=title,
            body_template=body_template,
            context_name=context_name,
            context_kwargs=context_kwargs,
            send_after=send_after,
            subject_template=subject_template,
            preheader_template=preheader_template,
            status=NotificationStatus.PENDING_SEND.value,
            adapter_extra_parameters=adapter_extra_parameters,
        )
        self.notifications.append(notification)
        self._store_notifications()
        return notification

    def persist_notification_update(
        self, notification_id: int | str | uuid.UUID, update_data: UpdateNotificationKwargs
    ) -> Notification:
        notification = self.get_notification(notification_id)

        for key, value in update_data.items():
            setattr(notification, key, value)

        self._store_notifications()
        return notification

    def mark_pending_as_sent(self, notification_id: int | str | uuid.UUID) -> Notification:
        notification = self.get_notification(notification_id)
        notification.status = NotificationStatus.SENT.value
        self._store_notifications()
        return notification

    def mark_pending_as_failed(self, notification_id: int | str | uuid.UUID) -> Notification:
        notification = self.get_notification(notification_id)
        notification.status = NotificationStatus.FAILED.value
        self._store_notifications()
        return notification

    def mark_sent_as_read(self, notification_id: int | str | uuid.UUID) -> Notification:
        notification = self.get_notification(notification_id)
        notification.status = NotificationStatus.READ.value
        self._store_notifications()
        return notification

    def cancel_notification(self, notification_id: int | str | uuid.UUID) -> None:
        notification = self.get_notification(notification_id)
        self.notifications.remove(notification)
        self._store_notifications()

    def get_notification(
        self, notification_id: int | str | uuid.UUID, for_update=False
    ) -> Notification:
        try:
            return next(n for n in self.notifications if str(n.id) == str(notification_id))
        except StopIteration as e:
            raise NotificationNotFoundError("Notification not found") from e

    def filter_all_in_app_unread_notifications(
        self, user_id: int | str | uuid.UUID
    ) -> list[Notification]:
        notifications = [
            n
            for n in self.notifications
            if n.user_id == user_id
            and n.status == NotificationStatus.SENT.value
            and n.notification_type == NotificationTypes.IN_APP.value
        ]
        return notifications

    def filter_in_app_unread_notifications(
        self, user_id: int | str | uuid.UUID, page: int, page_size: int
    ) -> list[Notification]:
        return self.__paginate_notifications(
            self.filter_all_in_app_unread_notifications(user_id), page, page_size
        )

    def __paginate_notifications(
        self, notifications: list[Notification], page: int, page_size: int
    ):
        # page is 1-indexed
        return notifications[((page - 1) * page_size) : ((page - 1) * page_size) + page_size]

    def get_user_email_from_notification(self, notification_id: int | str | uuid.UUID) -> str:
        notification = self.get_notification(notification_id)
        return str(notification.context_kwargs.get("email", "testemail@example.com"))

    def store_context_used(
        self,
        notification_id: int | str | uuid.UUID,
        context: dict,
        adapter_import_str: str,
    ) -> None:
        notification = self.get_notification(notification_id)
        notification.context_used = context
        notification.adapter_used = adapter_import_str
        self._store_notifications()


class Config:
    def __init__(self, config_a: Decimal | None = None, config_b: datetime.datetime | None = None):
        self.config_a = config_a if config_a is not None else Decimal("1.0")
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        self.config_b = config_b if config_b is not None else now


class FakeFileBackendWithNonSerializableKWArgs(FakeFileBackend):
    config: Config

    def __init__(
        self, database_file_name: str = "notifications.json", config: Config | None = None
    ):
        super().__init__(database_file_name=database_file_name, config=config)

    def _store_notifications(self):
        assert self.config.config_a == Decimal("1.0")
        assert isinstance(self.config.config_b, datetime.datetime)
        super()._store_notifications()


class InvalidBackend:
    def __init__(self, *_args, **_kwargs):
        pass


class FakeAsyncIOFileBackend(AsyncIOBaseNotificationBackend):
    notifications: list[Notification]
    database_file_name: str

    def __init__(self, database_file_name: str = "notifications.json", **kwargs):
        super().__init__(database_file_name=database_file_name, **kwargs)
        self.database_file_name = database_file_name
        try:
            notifications_file = open(self.database_file_name, encoding="utf-8")
        except FileNotFoundError:
            self.notifications = []
            return
        try:
            self.notifications = [
                self._convert_json_to_notification(n) for n in json.load(notifications_file)
            ]
            notifications_file.close()
        except json.JSONDecodeError:
            self.notifications = []
            return

    async def clear(self):
        self.notifications = []
        try:
            os.remove(self.database_file_name)
        except FileNotFoundError:
            pass

    async def get_future_notifications(self, page: int, page_size: int) -> list[Notification]:
        return self.__paginate_notifications(
            await self.get_all_future_notifications(), page, page_size
        )

    async def get_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID, page: int, page_size: int
    ) -> list[Notification]:
        return self.__paginate_notifications(
            await self.get_all_future_notifications_from_user(user_id), page, page_size
        )

    async def get_all_future_notifications(self) -> list[Notification]:
        return [
            n
            for n in self.notifications
            if n.status == NotificationStatus.PENDING_SEND.value
            and (
                n.send_after is not None
                and n.send_after > datetime.datetime.now(tz=datetime.timezone.utc)
            )
        ]

    async def get_all_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID
    ) -> list[Notification]:
        return [
            n
            for n in self.notifications
            if n.status == NotificationStatus.PENDING_SEND.value
            and (
                n.send_after is not None
                and n.send_after > datetime.datetime.now(tz=datetime.timezone.utc)
            )
            and str(n.user_id) == str(user_id)
        ]

    async def get_all_pending_notifications(self) -> list[Notification]:
        return [
            n
            for n in self.notifications
            if n.status == NotificationStatus.PENDING_SEND.value
            and (
                n.send_after is None
                or n.send_after <= datetime.datetime.now(tz=datetime.timezone.utc)
            )
        ]

    def _convert_notification_to_json(self, notification: Notification) -> dict:
        return {
            "id": str(notification.id),
            "user_id": str(notification.user_id),
            "notification_type": notification.notification_type,
            "title": notification.title,
            "body_template": notification.body_template,
            "context_name": notification.context_name,
            "context_kwargs": notification.context_kwargs,
            "send_after": notification.send_after.isoformat() if notification.send_after else None,
            "subject_template": notification.subject_template,
            "preheader_template": notification.preheader_template,
            "status": notification.status,
            "context_used": notification.context_used,
        }

    def _convert_json_to_notification(self, notification: dict) -> Notification:
        return Notification(
            id=notification["id"],
            user_id=notification["user_id"],
            notification_type=notification["notification_type"],
            title=notification["title"],
            body_template=notification["body_template"],
            context_name=notification["context_name"],
            context_kwargs=notification["context_kwargs"],
            send_after=(
                datetime.datetime.fromisoformat(notification["send_after"])
                if notification["send_after"]
                else None
            ),
            subject_template=notification["subject_template"],
            preheader_template=notification["preheader_template"],
            status=notification["status"],
            context_used=notification.get("context_used"),
        )

    async def _store_notifications(self, lock: asyncio.Lock | None = None):
        if lock is not None:
            await lock.acquire()
        json_output_file = open(self.database_file_name, "w", encoding="utf-8")
        json.dump(
            [self._convert_notification_to_json(n) for n in self.notifications],
            json_output_file,
        )
        json_output_file.close()
        if lock is not None:
            lock.release()

    async def get_pending_notifications(self, page: int, page_size: int) -> list[Notification]:
        pending_notifications = await self.get_all_pending_notifications()
        return pending_notifications[
            ((page - 1) * page_size) : ((page - 1) * page_size) + page_size
        ]

    async def persist_notification(
        self,
        user_id: uuid.UUID | str | int,
        notification_type: str,
        title: str,
        body_template: str,
        context_name: str,
        context_kwargs: dict[str, uuid.UUID | str | int],
        send_after: datetime.datetime | None,
        subject_template: str,
        preheader_template: str,
        adapter_extra_parameters: dict | None = None,
        lock: asyncio.Lock | None = None,
    ) -> Notification:
        notification = Notification(
            id=str(uuid.uuid4()),
            user_id=user_id,
            notification_type=notification_type,
            title=title,
            body_template=body_template,
            context_name=context_name,
            context_kwargs=context_kwargs,
            send_after=send_after,
            subject_template=subject_template,
            preheader_template=preheader_template,
            status=NotificationStatus.PENDING_SEND.value,
            adapter_extra_parameters=adapter_extra_parameters,
        )
        self.notifications.append(notification)
        await self._store_notifications(lock)
        return notification

    async def persist_notification_update(
        self,
        notification_id: int | str | uuid.UUID,
        update_data: UpdateNotificationKwargs,
        lock: asyncio.Lock | None = None,
    ) -> Notification:
        notification = await self.get_notification(notification_id)

        for key, value in update_data.items():
            setattr(notification, key, value)

        await self._store_notifications(lock)
        return notification

    async def mark_pending_as_sent(
        self, notification_id: int | str | uuid.UUID, lock: asyncio.Lock | None = None
    ) -> Notification:
        notification = await self.get_notification(notification_id)
        notification.status = NotificationStatus.SENT.value
        await self._store_notifications(lock)
        return notification

    async def mark_pending_as_failed(
        self, notification_id: int | str | uuid.UUID, lock: asyncio.Lock | None = None
    ) -> Notification:
        notification = await self.get_notification(notification_id)
        notification.status = NotificationStatus.FAILED.value
        await self._store_notifications(lock)
        return notification

    async def mark_sent_as_read(
        self, notification_id: int | str | uuid.UUID, lock: asyncio.Lock | None = None
    ) -> Notification:
        notification = await self.get_notification(notification_id)
        notification.status = NotificationStatus.READ.value
        await self._store_notifications(lock)
        return notification

    async def cancel_notification(
        self, notification_id: int | str | uuid.UUID, lock: asyncio.Lock | None = None
    ) -> None:
        notification = await self.get_notification(notification_id)
        self.notifications.remove(notification)
        await self._store_notifications(lock)

    async def get_notification(
        self, notification_id: int | str | uuid.UUID, for_update=False
    ) -> Notification:
        try:
            return next(n for n in self.notifications if str(n.id) == str(notification_id))
        except StopIteration as e:
            raise NotificationNotFoundError("Notification not found") from e

    async def filter_all_in_app_unread_notifications(
        self, user_id: int | str | uuid.UUID
    ) -> list[Notification]:
        notifications = [
            n
            for n in self.notifications
            if n.user_id == user_id
            and n.status == NotificationStatus.SENT.value
            and n.notification_type == NotificationTypes.IN_APP.value
        ]
        return notifications

    async def filter_in_app_unread_notifications(
        self, user_id: int | str | uuid.UUID, page: int, page_size: int
    ) -> list[Notification]:
        in_app_unread_notifications = await self.filter_all_in_app_unread_notifications(user_id)
        return self.__paginate_notifications(in_app_unread_notifications, page, page_size)

    def __paginate_notifications(
        self, notifications: list[Notification], page: int, page_size: int
    ):
        return notifications[((page - 1) * page_size) : ((page - 1) * page_size) + page_size]

    async def get_user_email_from_notification(self, notification_id: int | str | uuid.UUID) -> str:
        notification = await self.get_notification(notification_id)
        return str(notification.context_kwargs.get("email", "testemail@example.com"))

    async def store_context_used(
        self,
        notification_id: int | str | uuid.UUID,
        context: dict,
        adapter_import_str: str,
        lock: asyncio.Lock | None = None,
    ) -> None:
        notification = await self.get_notification(notification_id)
        notification.context_used = context
        notification.adapter_used = adapter_import_str
        await self._store_notifications(lock)



from collections.abc import Iterable
from typing import Any, cast

from vintasend.app_settings import NotificationSettings
from vintasend.services.notification_adapters.asyncio_base import AsyncIOBaseNotificationAdapter
from vintasend.services.notification_adapters.base import BaseNotificationAdapter
from vintasend.services.notification_backends.asyncio_base import AsyncIOBaseNotificationBackend
from vintasend.services.notification_backends.base import BaseNotificationBackend
from vintasend.services.notification_template_renderers.base import BaseNotificationTemplateRenderer


def _import_class(import_string: str) -> Any:
    module_name, class_name = import_string.rsplit(".", 1)
    module = __import__(module_name, fromlist=[class_name])
    return getattr(module, class_name)


def get_asyncio_notification_adapter_cls(adapter_import_str: str) -> Any:
    try:
        adapter_cls = _import_class(adapter_import_str)
    except (ImportError, ModuleNotFoundError) as e:
        raise ValueError(
            f"Notifications Adapter Error: Could not import {adapter_import_str}"
        ) from e

    if not issubclass(adapter_cls, AsyncIOBaseNotificationAdapter):
        raise ValueError(
            f"Notifications Adapter Error: {adapter_import_str} is not a valid AsyncIO notification adapter"
        )

    return adapter_cls


def get_notification_adapter_cls(adapter_import_str: str) -> Any:
    try:
        adapter_cls = _import_class(adapter_import_str)
    except (ImportError, ModuleNotFoundError) as e:
        raise ValueError(
            f"Notifications Adapter Error: Could not import {adapter_import_str}"
        ) from e

    if not issubclass(adapter_cls, BaseNotificationAdapter):
        raise ValueError(
            f"Notifications Adapter Error: {adapter_import_str} is not a valid notification adapter"
        )

    return adapter_cls


def get_notification_adapters(
    adapters_imports_strs: Iterable[
        tuple[str | tuple[str, dict[str, Any]], str | tuple[str, dict[str, Any]]]
    ]
    | None,
    backend: str | None = None,
    backend_kwargs: dict | None = None,
    config: Any = None,
) -> list[BaseNotificationAdapter]:
    default_adapters = []
    app_settings = NotificationSettings(config)
    adapters_imports_strs_with_default = (
        adapters_imports_strs
        if adapters_imports_strs is not None
        else app_settings.NOTIFICATION_ADAPTERS
    )
    for adapter_import_str, template_renderer_import_str in adapters_imports_strs_with_default:
        adapter_kwargs: dict = {}
        if isinstance(adapter_import_str, tuple):
            adapter_import_str, adapter_kwargs = adapter_import_str

        try:
            adapter_cls = _import_class(adapter_import_str)
        except (ImportError, ModuleNotFoundError) as e:
            raise ValueError(
                f"Notifications Adapter Error: Could not import {adapter_import_str}"
            ) from e

        try:
            adapter = adapter_cls(
                template_renderer_import_str,
                backend if backend else app_settings.NOTIFICATION_BACKEND,
                backend_kwargs,
                config,
                **adapter_kwargs,
            )
        except Exception as e:  # noqa: BLE001
            raise ValueError(
                f"Notifications Adapter Error: Could not instantiate {adapter_import_str}"
            ) from e

        if not isinstance(adapter, BaseNotificationAdapter):
            raise ValueError(
                f"Notifications Adapter Error: {adapter_import_str} is not a valid notification adapter"
            )

        default_adapters.append(cast(BaseNotificationAdapter, adapter))
    return default_adapters


def get_asyncio_notification_adapters(
    adapters_imports_strs: Iterable[
        tuple[str | tuple[str, dict[str, Any]], str | tuple[str, dict[str, Any]]]
    ]
    | None,
    backend: str | None = None,
    backend_kwargs: dict | None = None,
    config: Any = None,
) -> list[AsyncIOBaseNotificationAdapter]:
    default_adapters = []
    app_settings = NotificationSettings(config)
    adapters_imports_strs_with_default = (
        adapters_imports_strs
        if adapters_imports_strs is not None
        else app_settings.NOTIFICATION_ADAPTERS
    )
    for adapter_import_str, template_renderer_import_str in adapters_imports_strs_with_default:
        adapter_kwargs: dict = {}
        if isinstance(adapter_import_str, tuple):
            adapter_import_str, adapter_kwargs = adapter_import_str
        try:
            adapter_cls = _import_class(adapter_import_str)
        except (ImportError, ModuleNotFoundError) as e:
            raise ValueError(
                f"Notifications Adapter Error: Could not import {adapter_import_str}"
            ) from e

        try:
            adapter = adapter_cls(
                template_renderer_import_str,
                backend if backend else app_settings.NOTIFICATION_BACKEND,
                backend_kwargs,
                config,
                **adapter_kwargs,
            )
        except Exception as e:  # noqa: BLE001
            raise ValueError(
                f"Notifications Adapter Error: Could not instantiate {adapter_import_str}"
            ) from e

        if not isinstance(adapter, AsyncIOBaseNotificationAdapter):
            raise ValueError(
                f"Notifications Adapter Error: {adapter_import_str} is not a valid notification adapter"
            )

        default_adapters.append(cast(AsyncIOBaseNotificationAdapter, adapter))
    return default_adapters


def get_notification_backend(
    backend_import_str: str | None, backend_kwargs: dict | None = None, config: Any = None
) -> BaseNotificationBackend:
    app_settings = NotificationSettings(config)
    backend_import_str_with_fallback = (
        backend_import_str
        if backend_import_str is not None
        else app_settings.NOTIFICATION_BACKEND
    )

    try:
        backend_cls = _import_class(backend_import_str_with_fallback)
    except (ImportError, ModuleNotFoundError) as e:
        raise ValueError(
            f"Notifications Backend Error: Could not import {backend_import_str_with_fallback}"
        ) from e

    try:
        backend = backend_cls(**backend_kwargs) if backend_kwargs else backend_cls()
    except Exception as e:  # noqa: BLE001
        raise ValueError(
            f"Notifications Backend Error: Could not instantiate {backend_import_str_with_fallback}"
        ) from e

    if not isinstance(backend, BaseNotificationBackend):
        raise ValueError(
            f"Notifications Backend Error: {backend_import_str_with_fallback} is not a valid notification backend"
        )
    return cast(BaseNotificationBackend, backend)


def get_asyncio_notification_backend(
    backend_import_str: str | None, backend_kwargs: dict | None = None, config: Any = None
) -> AsyncIOBaseNotificationBackend:
    app_settings = NotificationSettings(config)
    backend_import_str_with_fallback = (
        backend_import_str
        if backend_import_str is not None
        else app_settings.NOTIFICATION_BACKEND
    )

    try:
        backend_cls = _import_class(backend_import_str_with_fallback)
    except (ImportError, ModuleNotFoundError) as e:
        raise ValueError(
            f"Notifications Backend Error: Could not import {backend_import_str_with_fallback}"
        ) from e

    try:
        backend = backend_cls(**backend_kwargs) if backend_kwargs else backend_cls()
    except Exception as e:  # noqa: BLE001
        raise ValueError(
            f"Notifications Backend Error: Could not instantiate {backend_import_str_with_fallback}"
        ) from e

    if not isinstance(backend, AsyncIOBaseNotificationBackend):
        raise ValueError(
            f"Notifications Backend Error: {backend_import_str_with_fallback} is not a valid AsyncIO notification backend"
        )
    return cast(AsyncIOBaseNotificationBackend, backend)


def get_template_renderer(
    template_renderer_import_str: str | tuple[str, dict[str, Any]],
) -> BaseNotificationTemplateRenderer:
    template_renderer_kwargs: dict[str, Any] = {}
    if isinstance(template_renderer_import_str, tuple):
        template_renderer_import_str, template_renderer_kwargs = template_renderer_import_str

    try:
        template_renderer_cls = _import_class(template_renderer_import_str)
    except (ImportError, ModuleNotFoundError) as e:
        raise ValueError(
            f"Notifications Template Renderer Error: Could not import {template_renderer_import_str}"
        ) from e

    try:
        template_renderer = template_renderer_cls(**template_renderer_kwargs)
    except Exception as e:  # noqa: BLE001
        raise ValueError(
            f"Notifications Template Renderer Error: Could not instantiate {template_renderer_import_str}"
        ) from e

    if not isinstance(template_renderer, BaseNotificationTemplateRenderer):
        raise ValueError(
            f"Notifications Template Renderer Error: {template_renderer_import_str} is not a valid template renderer"
        )

    return cast(BaseNotificationTemplateRenderer, template_renderer)



class NotificationError(ValueError):
    pass


class NotificationContextGenerationError(NotificationError):
    pass


class NotificationUpdateError(NotificationError):
    pass


class NotificationMarkFailedError(NotificationUpdateError):
    pass


class NotificationMarkSentError(NotificationUpdateError):
    pass


class NotificationCancelError(NotificationError):
    pass


class NotificationNotFoundError(NotificationError):
    pass


class NotificationAlreadySentError(NotificationError):
    pass


class NotificationUserNotFoundError(NotificationError):
    pass


class NotificationTemplateRenderingError(NotificationError):
    pass


class NotificationPreheaderTemplateRenderingError(NotificationTemplateRenderingError):
    pass


class NotificationSubjectTemplateRenderingError(NotificationTemplateRenderingError):
    pass


class NotificationBodyTemplateRenderingError(NotificationTemplateRenderingError):
    pass


class NotificationSendError(NotificationError):
    pass



from types import MappingProxyType
from typing import Any, Generic, TypeVar, cast


class BaseSingletonMeta(type):
    _instances: MappingProxyType[Any, Any]


T = TypeVar('T', bound=BaseSingletonMeta)


class SingletonMeta(Generic[T], BaseSingletonMeta):
    _instances: MappingProxyType["SingletonMeta[T]", T] = MappingProxyType({})

    def __call__(cls, *args: Any, **kwargs: Any) -> T:
        if cls not in cls._instances:
            _instances: dict["SingletonMeta[T]", T] = dict(cls._instances)
            _instances[cls] = cast(T, super().__call__(*args, **kwargs))
            cls._instances = MappingProxyType(_instances)
        return cls._instances[cls]



import os
from typing import Any, TypedDict, cast

from vintasend.utils.singleton_utils import SingletonMeta


class NotificationSettingsDict(TypedDict):
    NOTIFICATION_ADAPTERS: list[tuple[str, str]]
    NOTIFICATION_BACKEND: str | None
    NOTIFICATION_MODEL: str | None
    NOTIFICATION_DEFAULT_BCC_EMAILS: list[str]
    NOTIFICATION_DEFAULT_BASE_URL_PROTOCOL: str
    NOTIFICATION_DEFAULT_BASE_URL_DOMAIN: str
    NOTIFICATION_DEFAULT_FROM_EMAIL: str 

DEFAULT_SETTINGS: NotificationSettingsDict = {
    "NOTIFICATION_ADAPTERS": [],
    "NOTIFICATION_BACKEND": None,
    "NOTIFICATION_MODEL": None,
    "NOTIFICATION_DEFAULT_BCC_EMAILS": [],
    "NOTIFICATION_DEFAULT_BASE_URL_PROTOCOL": "http",
    "NOTIFICATION_DEFAULT_BASE_URL_DOMAIN": "example.com",
    "NOTIFICATION_DEFAULT_FROM_EMAIL": "foo@examplo.com",
}


def is_django_installed():
    try:
        import django  # type: ignore # noqa # pylint: disable=unused-import
        from django.conf import settings  # type: ignore # noqa # pylint: disable=unused-import

        return True if settings.configured else False
    except ImportError:
        return False


def is_flask_installed():
    try:
        import flask  # type: ignore # noqa # pylint: disable=unused-import

        return True if flask.current_app else False
    except (ImportError, RuntimeError):
        return False


def is_fastapi_installed():
    try:
        import fastapi  # type: ignore # noqa # pylint: disable=unused-import

        return True
    except ImportError:
        return False


def detect_framework():
    if is_django_installed():
        return "Django"
    elif is_flask_installed():
        return "Flask"
    elif is_fastapi_installed():
        return "FastAPI"
    else:
        return "Unknown"


def get_setting_with_env_var_fallback(
    setting_name: str, 
    framework_value: Any | None = None, 
    default_settings: NotificationSettingsDict = DEFAULT_SETTINGS
):
    return os.getenv(
        setting_name,
        framework_value if framework_value else default_settings.get(setting_name, None)
    )


def get_django_setting(setting_name: str):
    from django.conf import settings

    DJANGO_DEFAULT_SETTINGS: NotificationSettingsDict = {
        **DEFAULT_SETTINGS,
        "NOTIFICATION_ADAPTERS": [
            (
                "vintasend_django.services.notification_adapters.django_email.DjangoEmailNotificationAdapter",
                "vintasend_django.services.notification_template_renderers.django_templated_email_renderer.DjangoTemplatedEmailRenderer",
            ),
        ],
        "NOTIFICATION_BACKEND": "vintasend_django.services.notification_backends.django_db_notification_backend.DjangoDbNotificationBackend",
        "NOTIFICATION_MODEL": "vintasend_django.models.Notification",
    }

    return get_setting_with_env_var_fallback(setting_name, getattr(settings, setting_name, None), DJANGO_DEFAULT_SETTINGS)


def get_flask_setting(setting_name: str):
    from flask import current_app  # type: ignore # noqa # pylint: disable=import-outside-toplevels

    FLASK_DEFAULT_SETTINGS: NotificationSettingsDict = {
        **DEFAULT_SETTINGS,
        "NOTIFICATION_ADAPTERS": [
            (
                "vintasend_flask_mail.services.notification_adapters.flask_mail.FlaskMailNotificationAdapter",
                "vintasend_jinja.services.notification_template_renderers.jinja_template_renderer.JinjaTemplatedEmailRenderer",
            ),
        ],
        "NOTIFICATION_BACKEND": "vintasend_sqlalchemy.services.notification_backends.sqlalchemy_notification_backend.SQLAlchemyNotificationBackend",
    }

    return get_setting_with_env_var_fallback(setting_name, current_app.config.get(setting_name, None), FLASK_DEFAULT_SETTINGS)

def get_fastapi_setting(setting_name: str, config: Any):
    FASTAPI_DEFAULT_SETTINGS: NotificationSettingsDict = {
        **DEFAULT_SETTINGS,
        "NOTIFICATION_ADAPTERS": [
            (
                "vintasend_fastapi_mail.services.notification_adapters.fastapi_mail.FastAPIMailNotificationAdapter",
                "vintasend_jinja.services.notification_template_renderers.jinja_template_renderer.JinjaTemplatedEmailRenderer",
            ),
        ],
        "NOTIFICATION_BACKEND": "vintasend_sqlalchemy.services.notification_backends.sqlalchemy_notification_backend.SQLAlchemyNotificationBackend",
    }
    return get_setting_with_env_var_fallback(setting_name, getattr(config, setting_name, None), FASTAPI_DEFAULT_SETTINGS)


def get_config(setting_name: str, config: Any = None):
    framework = detect_framework()
    if framework == "Django":
        return get_django_setting(setting_name)
    elif framework == "Flask":
        return get_flask_setting(setting_name)
    elif framework == "FastAPI":
        return get_fastapi_setting(setting_name, config)
    else:
        return {}


class NotificationSettings(metaclass=SingletonMeta):
    NOTIFICATION_ADAPTERS: list[tuple[str, str]]
    NOTIFICATION_BACKEND: str
    NOTIFICATION_MODEL: str | None
    NOTIFICATION_DEFAULT_BCC_EMAILS: list[str]
    NOTIFICATION_DEFAULT_BASE_URL_PROTOCOL: str
    NOTIFICATION_DEFAULT_BASE_URL_DOMAIN: str
    NOTIFICATION_DEFAULT_FROM_EMAIL: str

    def __init__(self, config: Any = None):
        self.NOTIFICATION_ADAPTERS = cast(list[tuple[str, str]], get_config("NOTIFICATION_ADAPTERS", config))
        self.NOTIFICATION_BACKEND = cast(str, get_config("NOTIFICATION_BACKEND", config))
        self.NOTIFICATION_MODEL = cast(str| None, get_config("NOTIFICATION_MODEL", config))
        self.NOTIFICATION_DEFAULT_BCC_EMAILS = cast(list[str], get_config("NOTIFICATION_DEFAULT_BCC_EMAILS", config))
        self.NOTIFICATION_DEFAULT_BASE_URL_PROTOCOL = cast(str, get_config("NOTIFICATION_DEFAULT_BASE_URL_PROTOCOL", config))
        self.NOTIFICATION_DEFAULT_BASE_URL_DOMAIN = cast(str, get_config("NOTIFICATION_DEFAULT_BASE_URL_DOMAIN", config))
        self.NOTIFICATION_DEFAULT_FROM_EMAIL = cast(str, get_config("NOTIFICATION_DEFAULT_FROM_EMAIL", config))

    def get_notification_model_cls(self):
        if self.NOTIFICATION_MODEL is None:
            raise ValueError("NOTIFICATION_MODEL is not set in the settings.")
        module_name, class_name = self.NOTIFICATION_MODEL.rsplit(".", 1)
        module = __import__(module_name, fromlist=[class_name])
        return getattr(module, class_name)



import asyncio
import datetime
import uuid
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Iterable

from vintasend.services.utils import get_class_path


if TYPE_CHECKING:
    from vintasend.services.dataclasses import Notification, UpdateNotificationKwargs

class AsyncIOBaseNotificationBackend(ABC):
    def __init__(self, *args, **kwargs):
        self.backend_import_str = get_class_path(self)
        self.config = kwargs.pop("config", None)
        self.backend_kwargs = kwargs

    @abstractmethod
    async def get_all_pending_notifications(self) -> Iterable["Notification"]:
        ...

    @abstractmethod
    async def get_pending_notifications(
        self, page: int, page_size: int
    ) -> Iterable["Notification"]:
        ...

    @abstractmethod
    async def get_all_future_notifications(self) -> Iterable["Notification"]:
        ...

    @abstractmethod
    async def get_future_notifications(
        self, page: int, page_size: int
    ) -> Iterable["Notification"]:
        ...

    @abstractmethod
    async def get_all_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID
    ) -> Iterable["Notification"]:
        ...

    @abstractmethod
    async def get_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID, page: int, page_size: int
    ) -> Iterable["Notification"]:
        ...

    @abstractmethod
    async def persist_notification(
        self,
        user_id: int | str | uuid.UUID,
        notification_type: str,
        title: str,
        body_template: str,
        context_name: str,
        context_kwargs: dict[str, uuid.UUID | str | int],
        send_after: datetime.datetime | None,
        subject_template: str,
        preheader_template: str,
        adapter_extra_parameters: dict | None = None,
        lock: asyncio.Lock | None = None
    ) -> "Notification":
        ...

    @abstractmethod
    async def persist_notification_update(
        self,
        notification_id: int | str | uuid.UUID,
        update_data: "UpdateNotificationKwargs",
        lock: asyncio.Lock | None = None,
    ) -> "Notification":
        """
        Update a notification in the backend. This method should return the updated notification.
        Notifications that have already been sent should not be updated. 
        If a notification has already been sent, the method should raise a NotificationUpdateError.
        """
        ...

    @abstractmethod
    async def mark_pending_as_sent(
        self, notification_id: int | str | uuid.UUID, lock: asyncio.Lock | None = None
    ) -> "Notification":
        ...

    @abstractmethod
    async def mark_pending_as_failed(
        self, notification_id: int | str | uuid.UUID, lock: asyncio.Lock | None = None
    ) -> "Notification":
        ...

    @abstractmethod
    async def mark_sent_as_read(
        self, notification_id: int | str | uuid.UUID, lock: asyncio.Lock | None = None
    ) -> "Notification":
        ...

    @abstractmethod
    async def cancel_notification(
        self, notification_id: int | str | uuid.UUID, lock: asyncio.Lock | None = None
    ) -> None:
        ...

    @abstractmethod
    async def get_notification(
        self, notification_id: int | str | uuid.UUID, for_update=False
    ) -> "Notification":
        ...

    @abstractmethod
    async def filter_all_in_app_unread_notifications(
        self, user_id: int | str | uuid.UUID
    ) -> Iterable["Notification"]:
        ...

    @abstractmethod
    async def filter_in_app_unread_notifications(
        self, user_id: int | str | uuid.UUID, page: int, page_size: int
    ) -> Iterable["Notification"]:
        ...

    @abstractmethod
    async def get_user_email_from_notification(
        self, notification_id: int | str | uuid.UUID
    ) -> str:
        ...
    
    @abstractmethod
    async def store_context_used(
        self, 
        notification_id: int | str | uuid.UUID, 
        context: dict,
        adapter_import_str: str, 
        lock: asyncio.Lock | None = None,
    ) -> None:
        ...
    


from vintasend.exceptions import NotificationBodyTemplateRenderingError
from vintasend.services.notification_template_renderers.base_templated_email_renderer import (
    BaseTemplatedEmailRenderer,
    TemplatedEmail,
)


class FakeTemplateRenderer(BaseTemplatedEmailRenderer):
    def render(self, notification, context):
        return TemplatedEmail(
            subject=notification.subject_template, body=notification.body_template
        )


class FakeTemplateRendererWithException(BaseTemplatedEmailRenderer):
    def render(self, notification, context):
        raise NotificationBodyTemplateRenderingError("Fake error")


class InvalidTemplateRenderer:
    pass


class FakeTemplateRendererWithExceptionOnInit(FakeTemplateRendererWithException):
    def __init__(self):
        raise NotificationBodyTemplateRenderingError("Fake error")



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast, overload

from vintasend.services.notification_backends.base import BaseNotificationBackend
from vintasend.services.notification_template_renderers.base import BaseNotificationTemplateRenderer


if TYPE_CHECKING:
    from vintasend.constants import NotificationTypes
    from vintasend.services.dataclasses import Notification
    from vintasend.services.notification_service import NotificationContextDict


B = TypeVar("B", bound=BaseNotificationBackend)
T = TypeVar("T", bound=BaseNotificationTemplateRenderer)


class BaseNotificationAdapter(Generic[B, T], ABC):
    """
    Base class for notification adapters. All notification adapters should inherit from this class.

    The notification adapter is responsible for sending notifications to the user and also for
    marking them as sent or failed.
    """

    notification_type: "NotificationTypes"

    backend: B
    template_renderer: T
    adapter_import_str: str
    adapter_kwargs: dict

    @overload
    def __init__(
        self,
        template_renderer: T,
        backend: B,
        backend_kwargs: None = None,
        config: Any = None,
        **kwargs,
    ) -> None: ...

    @overload
    def __init__(
        self,
        template_renderer: str | tuple[str, dict[str, Any]],
        backend: str,
        backend_kwargs: dict | None = None,
        config: Any = None,
        **kwargs,
    ) -> None: ...

    @overload
    def __init__(
        self,
        template_renderer: str | tuple[str, dict[str, Any]],
        backend: B,
        backend_kwargs: None = None,
        config: Any = None,
        **kwargs,
    ) -> None: ...

    @overload
    def __init__(
        self,
        template_renderer: T,
        backend: str,
        backend_kwargs: dict | None = None,
        config: Any = None,
        **kwargs,
    ) -> None: ...

    def __init__(
        self,
        template_renderer: T | str | tuple[str, dict[str, Any]],
        backend: B | str | None,
        backend_kwargs: dict | None = None,
        config: Any = None,
        **kwargs,
    ) -> None:
        """
        Initialize the notification adapter.

        :param template_renderer: The template renderer to use to render the notification templates.
        :param backend: The backend to use to persist the notifications.
        :param backend_kwargs: The backend kwargs to pass to the backend in case backend is an import string.
        """
        from vintasend.services.helpers import get_notification_backend, get_template_renderer

        self.adapter_kwargs = kwargs

        if backend is not None and isinstance(backend, BaseNotificationBackend):
            self.backend = cast(B, backend)
        else:
            self.backend = cast(B, get_notification_backend(backend, backend_kwargs, config))
        
        if isinstance(template_renderer, str) or isinstance(template_renderer, tuple):
            self.template_renderer = cast(T, get_template_renderer(template_renderer))
        else:
            self.template_renderer = template_renderer

        self.adapter_import_str = f"{self.__module__}.{self.__class__.__name__}"
        self.config = config

    @abstractmethod
    def send(self, notification: "Notification", context: "NotificationContextDict") -> None:
        """
        Send the notification to the user.

        :param notification: The notification to send.
        :param context: The context to render the notification templates.
        """
        raise NotImplementedError



from enum import Enum


class NotificationStatus(Enum):
    PENDING_SEND = "PENDING_SEND"
    SENT = "SENT"
    FAILED = "FAILED"
    READ = "READ"
    CANCELLED = "CANCELLED"


class NotificationTypes(Enum):
    PUSH = "PUSH"
    EMAIL = "EMAIL"
    SMS = "SMS"
    IN_APP = "IN_APP"


from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from vintasend.services.utils import get_class_path


if TYPE_CHECKING:
    from vintasend.services.dataclasses import Notification
    from vintasend.services.notification_service import NotificationContextDict


class NotificationSendInput:
    """
    Represents the input necessary for the send method.
    """

    pass


class BaseNotificationTemplateRenderer(ABC):
    """
    Base class for notification template renderers. All notification template renderers should inherit from this class.

    The notification template renderer is responsible for rendering the notification templates.
    """

    template_renderer_import_str: str
    template_renderer_kwargs: dict

    def __init__(self, **kwargs):
        self.template_renderer_import_str = get_class_path(self)
        self.template_renderer_kwargs = kwargs

    @abstractmethod
    def render(
        self, notification: "Notification", context: "NotificationContextDict"
    ) -> NotificationSendInput:
        """
        Render the notification template.

        :param notification: The notification to render.
        :return: The input necessary to send the notification.
        """
        raise NotImplementedError



import logging
from typing import Any

from vintasend.services.helpers import (
    get_notification_adapter_cls,
)
from vintasend.services.notification_adapters.async_base import (
    AsyncBaseNotificationAdapter,
    NotificationDict,
)
from vintasend.services.notification_service import NotificationService


logger = logging.getLogger(__name__)


def send_notification(
    notification: NotificationDict,
    context: dict,
    adapters: list[tuple[str | tuple[str, dict[str, Any]], str | tuple[str, dict[str, Any]]]],
    backend: str,
    backend_kwargs: dict | None = None,
    config: dict | None = None,
):
    # Get the adapter class, only a single adapter is supported
    adapter_cls = get_notification_adapter_cls(
        adapters[0][0] if isinstance(adapters[0][0], str) else adapters[0][0][0]
    )
    
    if not issubclass(adapter_cls, AsyncBaseNotificationAdapter):
        return
    
    desserialized_backend_kwargs = (
        adapter_cls.restore_backend_kwargs(backend_kwargs) if backend_kwargs else None
    )
    desserialized_config = (
        adapter_cls.restore_config(config) if config else None
    )
    desserialized_adapter_kwargs = (
        adapter_cls.restore_adapter_kwargs(adapters[0][0][1]) if isinstance(adapters[0][0], tuple) else {}
    )

    desserialized_template_renderer_kwargs = (
        adapter_cls.restore_template_renderer_kwargs(adapters[0][1][1])
        if isinstance(adapters[0][1], tuple)
        else {}
    )

    adapters_import_tuple = (
        (adapters[0][0][0], desserialized_adapter_kwargs), 
        (adapters[0][1][0], desserialized_template_renderer_kwargs)
    )
    try:
        service: NotificationService[Any, Any] = NotificationService(
            [adapters_import_tuple], backend, desserialized_backend_kwargs, desserialized_config
        )
        service.delayed_send(notification, context)
    except Exception as e:
        logger.exception("Error sending notification: %s", e)



import asyncio
import datetime
import logging
import uuid
from collections.abc import Callable, Iterable
from typing import Any, ClassVar, Coroutine, Generic, TypeGuard, TypeVar, cast

from vintasend.app_settings import NotificationSettings
from vintasend.services.notification_backends.asyncio_base import AsyncIOBaseNotificationBackend


try:
    from typing import Unpack
except ImportError:
    from typing_extensions import Unpack

from vintasend.constants import NotificationTypes
from vintasend.exceptions import (
    NotificationContextGenerationError,
    NotificationError,
    NotificationMarkFailedError,
    NotificationMarkSentError,
    NotificationSendError,
    NotificationUpdateError,
)
from vintasend.services.dataclasses import (
    Notification,
    NotificationContextDict,
    UpdateNotificationKwargs,
)
from vintasend.services.helpers import (
    get_asyncio_notification_adapters,
    get_asyncio_notification_backend,
    get_notification_adapters,
    get_notification_backend,
)
from vintasend.services.notification_adapters.async_base import (
    AsyncBaseNotificationAdapter,
    NotificationDict,
)
from vintasend.services.notification_adapters.asyncio_base import AsyncIOBaseNotificationAdapter
from vintasend.services.notification_adapters.base import BaseNotificationAdapter
from vintasend.services.notification_backends.base import BaseNotificationBackend
from vintasend.services.utils import get_class_path
from vintasend.utils.singleton_utils import SingletonMeta


logger = logging.getLogger(__name__)


class Contexts(metaclass=SingletonMeta):
    _contexts: ClassVar[
        dict[
            str,
            Callable[[Any], NotificationContextDict]
            | Callable[[Any], Coroutine[Any, Any, NotificationContextDict]],
        ]
    ] = {}

    def register_function(
        self,
        key: str,
        func: Callable[[Any], NotificationContextDict]
        | Callable[[Any], Coroutine[Any, Any, NotificationContextDict]],
    ):
        self._contexts[key] = func

    def get_function(self, key: str):
        return self._contexts.get(key)


def register_context(key: str):
    def decorator(func: Callable[[Any], NotificationContextDict]):
        contexts = Contexts()
        contexts.register_function(key, func)
        return func

    return decorator


A = TypeVar("A", BaseNotificationAdapter, AsyncBaseNotificationAdapter)
B = TypeVar("B", bound=BaseNotificationBackend)


class NotificationService(Generic[A, B]):
    notification_adapters: Iterable[A]
    notification_backend: B

    def __init__(
        self,
        notification_adapters: Iterable[A]
        | Iterable[tuple[str, str | tuple[str, dict[str, Any]]]]
        | None = None,
        notification_backend: B | str | None = None,
        notification_backend_kwargs: dict | None = None,
        config: Any = None,
    ):
        # initialize the notification settings singleton for the first time
        # to ensure all components have access to the same settings
        NotificationSettings(config)

        if isinstance(notification_backend, BaseNotificationBackend):
            self.notification_backend = cast(B, notification_backend)
        else:
            self.notification_backend = cast(
                B,
                get_notification_backend(notification_backend, notification_backend_kwargs, config),
            )
        self.notification_backend_import_str = get_class_path(self.notification_backend)

        if notification_adapters is None or self._check_is_adapters_tuple_iterable(
            notification_adapters
        ):
            self.notification_adapters = cast(
                Iterable[A],
                get_notification_adapters(
                    notification_adapters,
                    self.notification_backend_import_str,
                    notification_backend_kwargs if notification_backend_kwargs is not None else {},
                    config,
                ),
            )
        elif self._check_is_base_notification_adapter_iterable(notification_adapters):
            self.notification_adapters = notification_adapters
        else:
            raise NotificationError("Invalid notification adapters")
        self.notification_adapters_import_strs = [
            (get_class_path(adapter), get_class_path(adapter.template_renderer))
            for adapter in self.notification_adapters
        ]

    def _check_is_base_notification_adapter_iterable(
        self,
        notification_adapters: Iterable[A]
        | Iterable[tuple[str, str | tuple[str, dict[str, Any]]]]
        | None,
    ) -> TypeGuard[Iterable[A]]:
        return notification_adapters is not None and all(
            isinstance(adapter, BaseNotificationAdapter) for adapter in notification_adapters
        )

    def _check_is_adapters_tuple_iterable(
        self,
        notification_adapters: Iterable[A]
        | Iterable[tuple[str, str | tuple[str, dict[str, Any]]]]
        | None,
    ) -> TypeGuard[Iterable[tuple[str, str | tuple[str, dict[str, Any]]]]]:
        return notification_adapters is not None and all(
            (isinstance(adapter, tuple) or isinstance(adapter, list))
            and len(adapter) == 2
            and isinstance(adapter[0], str)
            and (
                isinstance(adapter[1], str)
                or (
                    isinstance(adapter[1], tuple)
                    and isinstance(adapter[1][0], str)
                    and isinstance(adapter[1][1], dict)
                )
            )
            for adapter in notification_adapters
        )

    def send(self, notification: Notification) -> None:
        """
        Send a notification using the appropriate adapter

        This method may raise the following exceptions:
            * NotificationUserNotFoundError if the user for the notification can't be found;
            * NotificationContextGenerationError if the context generation fails;
            * NotificationSendError if the adapter fails to send the notification.
            * NotificationMarkFailedError if the notification fails to be marked as failed.
            * NotificationMarkSentError if the notification fails to be marked as sent.

        Parameters:
            notification: Notification - the notification to be sent
        """
        try:
            context = self.get_notification_context(notification)
        except NotificationContextGenerationError:
            logger.exception("Failed to generate context for notification %s", notification.id)
            try:
                self.notification_backend.mark_pending_as_failed(notification.id)
            except NotificationUpdateError as e:
                raise NotificationMarkFailedError("Failed to mark notification as failed") from e
            return

        for adapter in self.notification_adapters:
            if adapter.notification_type.value != notification.notification_type:
                continue
            try:
                adapter.send(
                    notification=notification,
                    context=context,
                )
                if isinstance(adapter, AsyncBaseNotificationAdapter):
                    return
            except Exception as e:  # noqa: BLE001
                try:
                    raise NotificationSendError("Failed to send notification") from e
                except NotificationSendError as e:
                    try:
                        self.notification_backend.mark_pending_as_failed(notification.id)
                    except NotificationUpdateError:
                        raise NotificationMarkFailedError(
                            "Failed to mark notification as failed"
                        ) from e
                    raise e
            try:
                self.notification_backend.mark_pending_as_sent(notification.id)
                self.notification_backend.store_context_used(
                    notification.id, 
                    context,
                    adapter.adapter_import_str,
                )
            except NotificationUpdateError as e:
                raise NotificationMarkSentError("Failed to mark notification as sent") from e

    def create_notification(
        self,
        user_id: int | str | uuid.UUID,
        notification_type: str,
        title: str,
        body_template: str,
        context_name: str,
        context_kwargs: NotificationContextDict,
        send_after: datetime.datetime | None = None,
        subject_template: str = "",
        preheader_template: str = "",
        adapter_extra_parameters: dict | None = None,
    ) -> Notification:
        """
        Create a notification and send it if it is due to be sent immediately.

        This method may raise the following exceptions:
            * NotificationContextGenerationError if the context generation fails;
            * NotificationSendError if the adapter fails to send the notification.
            * NotificationMarkFailedError if the notification fails to be marked as failed.
            * NotificationMarkSentError if the notification fails to be marked as sent.

        Parameters:
            user_id: int | str | uuid.UUID - the user ID to send the notification to
            notification_type: str - the type of notification to send
            title: str - the title of the notification
            body_template: str - the string that represents the body template
            context_name: str - the name of the context function to generate the context
            context_kwargs: NotificationContextDict - the context kwargs to generate the context
            send_after: datetime.datetime | None - the date and time to send the notification
            subject_template: str - the  string that represents the subject template
            preheader_template: str - the string that represents the preheader template
        """
        notification = self.notification_backend.persist_notification(
            user_id=user_id,
            notification_type=notification_type,
            title=title,
            body_template=body_template,
            context_name=context_name,
            context_kwargs=context_kwargs,
            send_after=send_after,
            subject_template=subject_template,
            preheader_template=preheader_template,
            adapter_extra_parameters=adapter_extra_parameters,
        )
        if notification.send_after is None or notification.send_after <= datetime.datetime.now(
            tz=datetime.timezone.utc
        ):
            self.send(notification)
        return notification

    def update_notification(
        self,
        notification_id: int | str | uuid.UUID,
        **kwargs: Unpack[UpdateNotificationKwargs],
    ) -> Notification:
        """
        Update a notification and send it if it is due to be sent immediately.

        This method may raise the following exceptions:
            * NotificationContextGenerationError if the context generation fails;
            * NotificationSendError if the adapter fails to send the notification.
            * NotificationMarkFailedError if the notification fails to be marked as failed.
            * NotificationMarkSentError if the notification fails to be marked as sent.

        Parameters:
            notification_id: int | str | uuid.UUID - the ID of the notification to update
            **kwargs: UpdateNotificationKwargs - the fields to update
        """
        notification = self.notification_backend.persist_notification_update(
            notification_id=notification_id,
            update_data=kwargs,
        )
        if notification.send_after is None or notification.send_after <= datetime.datetime.now(
            tz=datetime.timezone.utc
        ):
            self.send(notification)
        return notification

    def get_all_future_notifications(self) -> Iterable[Notification]:
        """
        Get future notifications from the backend.

        Returns:
            Iterable[Notification] - the future notifications
        """
        return self.notification_backend.get_all_future_notifications()

    def get_all_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID
    ) -> Iterable[Notification]:
        """
        Get future notifications from the backend.

        Parameters:
            user_id: int | str | uuid.UUID - the user ID to get the notifications for

        Returns:
            Iterable[Notification] - the future notifications from the user
        """
        return self.notification_backend.get_all_future_notifications_from_user(user_id)

    def get_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID, page: int, page_size: int
    ) -> Iterable[Notification]:
        """
        Get future notifications from the backend.

        Parameters:
            user_id: int | str | uuid.UUID - the user ID to get the notifications for
            page: int - the page number to get
            page_size: int - the number of notifications per page

        Returns:
            Iterable[Notification] - the selected page of the future notifications from the user
        """
        return self.notification_backend.get_future_notifications_from_user(
            user_id, page, page_size
        )

    def get_future_notifications(self, page: int, page_size: int) -> Iterable[Notification]:
        """
        Get future notifications from the backend.

        Parameters:
            user_id: int | str | uuid.UUID - the user ID to get the notifications for
            page: int - the page number to get
            page_size: int - the number of notifications per page

        Returns:
            Iterable[Notification] - the future notifications
        """
        return self.notification_backend.get_future_notifications(page, page_size)

    def _is_asyncio_context_function(
        self,
        context_function: Callable[[Any], NotificationContextDict]
        | Callable[[Any], Coroutine[Any, Any, NotificationContextDict]],
    ) -> TypeGuard[Callable[[Any], Coroutine[Any, Any, NotificationContextDict]]]:
        return asyncio.iscoroutinefunction(context_function)

    def _is_sync_context_function(
        self,
        context_function: Callable[[Any], NotificationContextDict]
        | Callable[[Any], Coroutine[Any, Any, NotificationContextDict]],
    ) -> TypeGuard[Callable[[Any], NotificationContextDict]]:
        return not asyncio.iscoroutinefunction(context_function)

    def get_notification_context(self, notification: Notification) -> NotificationContextDict:
        """
        Generate the context for a notification. It uses the context_name and context_kwargs from the notification.
        Contexts are registered using the @register_context decorator.

        This method may raise the following exceptions:
            * NotificationContextGenerationError if the context generation fails.

        Parameters:
            notification: Notification - the notification to generate the context for
        """
        context_function = Contexts().get_function(notification.context_name)
        if context_function is None:
            raise NotificationContextGenerationError("Context function not found")
        try:
            if self._is_asyncio_context_function(context_function):
                return asyncio.run(context_function(*[], **notification.context_kwargs))
            elif self._is_sync_context_function(context_function):
                return context_function(*[], **notification.context_kwargs)
            raise NotificationContextGenerationError("Invalid context function")
        except Exception as e:  # noqa: BLE001
            raise NotificationContextGenerationError("Failed getting notification context") from e

    def send_pending_notifications(self) -> None:
        """
        Send all pending notifications in the backend.

        This method doesn't raise any exceptions, but it provides specific logs for each
        notification (success or failure) and a summary at the end with the number of notifications
        sent and failed.
        """

        pending_notifications = self.notification_backend.get_all_pending_notifications()
        notifications_sent = 0
        notifications_failed = 0
        for notification in pending_notifications:
            try:
                self.send(notification)
            except NotificationSendError:
                notifications_failed += 1
                logger.exception("Failed to send notification %s", notification.id)
            except NotificationMarkFailedError:
                notifications_failed += 1
                logger.exception("Failed to send notification %s", notification.id)
                logger.exception("Failed to mark notification %s as failed", notification.id)
            except NotificationMarkSentError:
                logger.info("Notification %s sent", notification.id)
                logger.exception("Failed to mark notification %s as sent", notification.id)
                notifications_sent += 1
            else:
                logger.info("Notification %s sent", notification.id)
                notifications_sent += 1

        logger.info("Sent %s notifications", notifications_sent)
        logger.info("Failed to send %s notifications", notifications_failed)

    def get_pending_notifications(self, page: int, page_size: int) -> Iterable[Notification]:
        """
        Get pending notifications from the backend.

        Parameters:
            page: int - the page number to get
            page_size: int - the number of notifications per page

        Returns:
            Iterable[Notification] - the pending notifications
        """
        return self.notification_backend.get_pending_notifications(page, page_size)

    def get_notification(self, notification_id: int | str | uuid.UUID) -> Notification:
        """
        Get a notification from the backend.

        Parameters:
            notification_id: int | str | uuid.UUID - the ID of the notification to get

        Returns:
            Notification - the notification
        """
        return self.notification_backend.get_notification(notification_id)

    def mark_read(self, notification_id: int | str | uuid.UUID) -> Notification:
        """
        Mark a notification as read.

        This method may raise the following exceptions:
            * NotificationUpdateError if the notification fails to be marked as read.

        Parameters:
            notification: Notification - the notification to mark as read

        Returns:
            Notification - the updated notification
        """
        return self.notification_backend.mark_sent_as_read(notification_id)

    def get_in_app_unread(
        self,
        user_id: int | str | uuid.UUID,
        page: int = 1,
        page_size: int = 10,
    ) -> Iterable[Notification]:
        """
        Get unread in-app notifications for a user.

        This method may raise the following exceptions:
            * NotificationError if no in-app notification adapter is found.

        Parameters:
            user_id: int | str | uuid.UUID - the user ID to get the notifications for
            page: int - the page number to get
            page_size: int - the number of notifications per page

        Returns:
            Iterable[Notification] - the unread in-app notifications
        """
        if not any(
            a.notification_type == NotificationTypes.IN_APP for a in self.notification_adapters
        ):
            raise NotificationError("No in-app notification adapter found")
        return self.notification_backend.filter_in_app_unread_notifications(
            user_id=user_id, page=page, page_size=page_size
        )

    def cancel_notification(self, notification_id: int | str | uuid.UUID) -> None:
        """
        Cancel a notification.

        Parameters:
            notifictaion_id: int | str | uuid.UUID - the ID of the notification to cancel
        """
        return self.notification_backend.cancel_notification(notification_id)

    def delayed_send(self, notification_dict: NotificationDict, context_dict: dict) -> None:
        """
        Send a notification using the appropriate adapter with a delay.

        This method may raise the following exceptions:
            * NotificationUserNotFoundError if the user for the notification can't be found;
            * NotificationContextGenerationError if the context generation fails;
            * NotificationSendError if the adapter fails to send the notification.
            * NotificationMarkFailedError if the notification fails to be marked as failed.
            * NotificationMarkSentError if the notification fails to be marked as sent.

        Parameters:
            notification_dict: dict - the notification to be sent
            context_dict: dict - the context to generate the context for the notification
        """
        for adapter in self.notification_adapters:
            if adapter.notification_type.value != notification_dict.get("notification_type"):
                continue

            if not isinstance(adapter, AsyncBaseNotificationAdapter):
                return None

            async_adapter = cast(AsyncBaseNotificationAdapter, adapter)
            try:
                async_adapter.delayed_send(
                    notification_dict=notification_dict, context_dict=context_dict
                )
            except Exception as e:  # noqa: BLE001
                try:
                    raise NotificationSendError("Failed to send notification") from e
                except NotificationSendError as e:
                    try:
                        self.notification_backend.mark_pending_as_failed(notification_dict["id"])
                    except NotificationUpdateError:
                        raise NotificationMarkFailedError(
                            "Failed to mark notification as failed"
                        ) from e
                    raise e
            try:
                self.notification_backend.mark_pending_as_sent(notification_dict["id"])
                self.notification_backend.store_context_used(
                    notification_dict["id"], 
                    context_dict,
                    async_adapter.adapter_import_str,
                )
            except NotificationUpdateError as e:
                raise NotificationMarkSentError("Failed to mark notification as sent") from e


AAIO = TypeVar("AAIO", bound=AsyncIOBaseNotificationAdapter)
BAIO = TypeVar("BAIO", bound=AsyncIOBaseNotificationBackend)


class AsyncIONotificationService(Generic[AAIO, BAIO]):
    notification_adapters: Iterable[AAIO]
    notification_backend: BAIO

    def __init__(
        self,
        notification_adapters: Iterable[AAIO] | Iterable[tuple[str, str]] | None = None,
        notification_backend: BAIO | str | None = None,
        notification_backend_kwargs: dict | None = None,
        config: Any = None,
    ):
        if isinstance(notification_backend, AsyncIOBaseNotificationBackend):
            self.notification_backend = cast(BAIO, notification_backend)
        else:
            self.notification_backend = cast(
                BAIO,
                get_asyncio_notification_backend(
                    notification_backend, notification_backend_kwargs, config
                ),
            )
        self.notification_backend_import_str = get_class_path(self.notification_backend)

        if notification_adapters is None or self._check_is_adapters_tuple_iterable(
            notification_adapters
        ):
            self.notification_adapters = cast(
                Iterable[AAIO],
                get_asyncio_notification_adapters(
                    notification_adapters,
                    self.notification_backend_import_str,
                    notification_backend_kwargs if notification_backend_kwargs is not None else {},
                    config,
                ),
            )
        elif self._check_is_base_notification_adapter_iterable(notification_adapters):
            self.notification_adapters = notification_adapters
        else:
            raise NotificationError("Invalid notification adapters")
        self.notification_adapters_import_strs = [
            (get_class_path(adapter), get_class_path(adapter.template_renderer))
            for adapter in self.notification_adapters
        ]

    def _check_is_base_notification_adapter_iterable(
        self, notification_adapters: Iterable[AAIO] | Iterable[tuple[str, str]] | None
    ) -> TypeGuard[Iterable[AAIO]]:
        return notification_adapters is not None and all(
            isinstance(adapter, AsyncIOBaseNotificationAdapter) for adapter in notification_adapters
        )

    def _check_is_adapters_tuple_iterable(
        self, notification_adapters: Iterable[AAIO] | Iterable[tuple[str, str]] | None
    ) -> TypeGuard[Iterable[tuple[str, str]]]:
        return notification_adapters is not None and all(
            (isinstance(adapter, tuple) or isinstance(adapter, list))
            and len(adapter) == 2
            and isinstance(adapter[0], str)
            and isinstance(adapter[1], str)
            for adapter in notification_adapters
        )

    async def send(self, notification: Notification, lock: asyncio.Lock | None = None) -> None:
        """
        Send a notification using the appropriate adapter

        This method may raise the following exceptions:
            * NotificationUserNotFoundError if the user for the notification can't be found;
            * NotificationContextGenerationError if the context generation fails;
            * NotificationSendError if the adapter fails to send the notification.
            * NotificationMarkFailedError if the notification fails to be marked as failed.
            * NotificationMarkSentError if the notification fails to be marked as sent.

        Parameters:
            notification: Notification - the notification to be sent
        """
        try:
            context = await self.get_notification_context(notification)
        except NotificationContextGenerationError:
            logger.exception("Failed to generate context for notification %s", notification.id)
            try:
                await self.notification_backend.mark_pending_as_failed(notification.id, lock)
            except NotificationUpdateError as e:
                raise NotificationMarkFailedError("Failed to mark notification as failed") from e
            return None

        for adapter in self.notification_adapters:
            if adapter.notification_type.value != notification.notification_type:
                continue
            if not isinstance(adapter, AsyncIOBaseNotificationAdapter):
                continue
            try:
                await adapter.send(
                    notification=notification,
                    context=context,
                )
            except Exception as e:
                try:
                    raise NotificationSendError("Failed to send notification") from e
                except NotificationSendError as e:
                    try:
                        await self.notification_backend.mark_pending_as_failed(
                            notification.id, lock
                        )
                    except NotificationUpdateError:
                        raise NotificationMarkFailedError(
                            "Failed to mark notification as failed"
                        ) from e
                    raise e
            try:
                await self.notification_backend.mark_pending_as_sent(notification.id, lock)
                await self.notification_backend.store_context_used(
                    notification.id, 
                    context, 
                    adapter.adapter_import_str,
                    lock
                )
            except NotificationUpdateError as e:
                raise NotificationMarkSentError("Failed to mark notification as sent") from e
        return None

    async def create_notification(
        self,
        user_id: int | str | uuid.UUID,
        notification_type: str,
        title: str,
        body_template: str,
        context_name: str,
        context_kwargs: NotificationContextDict,
        send_after: datetime.datetime | None = None,
        subject_template: str = "",
        preheader_template: str = "",
        adapter_extra_parameters: dict | None = None,
    ) -> Notification:
        """
        Create a notification and send it if it is due to be sent immediately.

        This method may raise the following exceptions:
            * NotificationContextGenerationError if the context generation fails;
            * NotificationSendError if the adapter fails to send the notification.
            * NotificationMarkFailedError if the notification fails to be marked as failed.
            * NotificationMarkSentError if the notification fails to be marked as sent.

        Parameters:
            user_id: int | str | uuid.UUID - the user ID to send the notification to
            notification_type: str - the type of notification to send
            title: str - the title of the notification
            body_template: str - the string that represents the body template
            context_name: str - the name of the context function to generate the context
            context_kwargs: NotificationContextDict - the context kwargs to generate the context
            send_after: datetime.datetime | None - the date and time to send the notification
            subject_template: str - the  string that represents the subject template
            preheader_template: str - the string that represents the preheader template
        """
        notification = await self.notification_backend.persist_notification(
            user_id=user_id,
            notification_type=notification_type,
            title=title,
            body_template=body_template,
            context_name=context_name,
            context_kwargs=context_kwargs,
            send_after=send_after,
            subject_template=subject_template,
            preheader_template=preheader_template,
            adapter_extra_parameters=adapter_extra_parameters,
        )
        if notification.send_after is None or notification.send_after <= datetime.datetime.now(
            tz=datetime.timezone.utc
        ):
            await self.send(notification)
        return notification

    async def update_notification(
        self,
        notification_id: int | str | uuid.UUID,
        **kwargs: Unpack[UpdateNotificationKwargs],
    ) -> Notification:
        """
        Update a notification and send it if it is due to be sent immediately.

        This method may raise the following exceptions:
            * NotificationContextGenerationError if the context generation fails;
            * NotificationSendError if the adapter fails to send the notification.
            * NotificationMarkFailedError if the notification fails to be marked as failed.
            * NotificationMarkSentError if the notification fails to be marked as sent.

        Parameters:
            notification_id: int | str | uuid.UUID - the ID of the notification to update
            **kwargs: UpdateNotificationKwargs - the fields to update
        """
        notification = await self.notification_backend.persist_notification_update(
            notification_id=notification_id,
            update_data=kwargs,
        )
        if notification.send_after is None or notification.send_after <= datetime.datetime.now(
            tz=datetime.timezone.utc
        ):
            await self.send(notification)
        return notification

    async def get_all_future_notifications(self) -> Iterable[Notification]:
        """
        Get future notifications from the backend.

        Returns:
            Iterable[Notification] - the future notifications
        """
        return await self.notification_backend.get_all_future_notifications()

    async def get_all_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID
    ) -> Iterable[Notification]:
        """
        Get future notifications from the backend.

        Parameters:
            user_id: int | str | uuid.UUID - the user ID to get the notifications for

        Returns:
            Iterable[Notification] - the future notifications from the user
        """
        return await self.notification_backend.get_all_future_notifications_from_user(user_id)

    async def get_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID, page: int, page_size: int
    ) -> Iterable[Notification]:
        """
        Get future notifications from the backend.

        Parameters:
            user_id: int | str | uuid.UUID - the user ID to get the notifications for
            page: int - the page number to get
            page_size: int - the number of notifications per page

        Returns:
            Iterable[Notification] - the selected page of the future notifications from the user
        """
        return await self.notification_backend.get_future_notifications_from_user(
            user_id, page, page_size
        )

    async def get_future_notifications(self, page: int, page_size: int) -> Iterable[Notification]:
        """
        Get future notifications from the backend.

        Parameters:
            user_id: int | str | uuid.UUID - the user ID to get the notifications for
            page: int - the page number to get
            page_size: int - the number of notifications per page

        Returns:
            Iterable[Notification] - the future notifications
        """
        return await self.notification_backend.get_future_notifications(page, page_size)

    async def get_notification_context(self, notification: Notification) -> NotificationContextDict:
        """
        Generate the context for a notification. It uses the context_name and context_kwargs from the notification.
        Contexts are registered using the @register_context decorator.

        This method may raise the following exceptions:
            * NotificationContextGenerationError if the context generation fails.

        Parameters:
            notification: Notification - the notification to generate the context for
        """
        context_function = Contexts().get_function(notification.context_name)
        if context_function is None:
            raise NotificationContextGenerationError("Context function not found")
        try:
            if self._is_asyncio_context_function(context_function):
                return await context_function(*[], **notification.context_kwargs)
            elif self._is_sync_context_function(context_function):
                return context_function(*[], **notification.context_kwargs)
            raise NotificationContextGenerationError("Invalid context function")
        except Exception as e:  # noqa: BLE001
            raise NotificationContextGenerationError("Failed getting notification context") from e

    def _is_asyncio_context_function(
        self,
        context_function: Callable[[Any], NotificationContextDict]
        | Callable[[Any], Coroutine[Any, Any, NotificationContextDict]],
    ) -> TypeGuard[Callable[[Any], Coroutine[Any, Any, NotificationContextDict]]]:
        return asyncio.iscoroutinefunction(context_function)

    def _is_sync_context_function(
        self,
        context_function: Callable[[Any], NotificationContextDict]
        | Callable[[Any], Coroutine[Any, Any, NotificationContextDict]],
    ) -> TypeGuard[Callable[[Any], NotificationContextDict]]:
        return not asyncio.iscoroutinefunction(context_function)

    async def _send_notification_with_error_logging(
        self, notification: "Notification", lock: asyncio.Lock | None = None
    ) -> None:
        try:
            await self.send(notification, lock)
        except NotificationSendError:
            logger.exception("Failed to send notification %s", notification.id)
        except NotificationMarkFailedError:
            logger.exception("Failed to send notification %s", notification.id)
            logger.exception("Failed to mark notification %s as failed", notification.id)
        except NotificationMarkSentError:
            logger.info("Notification %s sent", notification.id)
            logger.exception("Failed to mark notification %s as sent", notification.id)
        else:
            logger.info("Notification %s sent", notification.id)

    async def send_pending_notifications(self) -> None:
        """
        Send all pending notifications in the backend.

        This method doesn't raise any exceptions, but it provides specific logs for each
        notification (success or failure) and a summary at the end with the number of notifications
        sent and failed.
        """

        pending_notifications = await self.notification_backend.get_all_pending_notifications()
        lock = asyncio.Lock()
        await asyncio.gather(
            *[
                self._send_notification_with_error_logging(notification, lock)
                for notification in pending_notifications
            ]
        )
        return None

    async def get_pending_notifications(self, page: int, page_size: int) -> Iterable[Notification]:
        """
        Get pending notifications from the backend.

        Parameters:
            page: int - the page number to get
            page_size: int - the number of notifications per page

        Returns:
            Iterable[Notification] - the pending notifications
        """
        return await self.notification_backend.get_pending_notifications(page, page_size)

    async def get_notification(self, notification_id: int | str | uuid.UUID) -> Notification:
        """
        Get a notification from the backend.

        Parameters:
            notification_id: int | str | uuid.UUID - the ID of the notification to get

        Returns:
            Notification - the notification
        """
        return await self.notification_backend.get_notification(notification_id)

    async def mark_read(self, notification_id: int | str | uuid.UUID) -> Notification:
        """
        Mark a notification as read.

        This method may raise the following exceptions:
            * NotificationUpdateError if the notification fails to be marked as read.

        Parameters:
            notification: Notification - the notification to mark as read

        Returns:
            Notification - the updated notification
        """
        return await self.notification_backend.mark_sent_as_read(notification_id)

    async def get_in_app_unread(
        self,
        user_id: int | str | uuid.UUID,
        page: int = 1,
        page_size: int = 10,
    ) -> Iterable[Notification]:
        """
        Get unread in-app notifications for a user.

        This method may raise the following exceptions:
            * NotificationError if no in-app notification adapter is found.

        Parameters:
            user_id: int | str | uuid.UUID - the user ID to get the notifications for
            page: int - the page number to get
            page_size: int - the number of notifications per page

        Returns:
            Iterable[Notification] - the unread in-app notifications
        """
        if not any(
            a.notification_type == NotificationTypes.IN_APP for a in self.notification_adapters
        ):
            raise NotificationError("No in-app notification adapter found")
        return await self.notification_backend.filter_in_app_unread_notifications(
            user_id=user_id, page=page, page_size=page_size
        )

    async def cancel_notification(self, notification_id: int | str | uuid.UUID) -> None:
        """
        Cancel a notification.

        Parameters:
            notifictaion_id: int | str | uuid.UUID - the ID of the notification to cancel
        """
        return await self.notification_backend.cancel_notification(notification_id)



import datetime
import uuid
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from vintasend.services.utils import get_class_path


if TYPE_CHECKING:
    from vintasend.services.dataclasses import (
        Notification,
        UpdateNotificationKwargs,
    )


class BaseNotificationBackend(ABC):
    backend_import_str: str
    backend_kwargs: dict
    config: Any

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        self.backend_import_str = get_class_path(self)
        self.config = kwargs.pop("config", None)
        self.backend_kwargs = kwargs

    @abstractmethod
    def get_all_pending_notifications(self) -> Iterable["Notification"]:
        raise NotImplementedError

    @abstractmethod
    def get_pending_notifications(self, page: int, page_size: int) -> Iterable["Notification"]:
        raise NotImplementedError

    @abstractmethod
    def get_all_future_notifications(self) -> Iterable["Notification"]:
        raise NotImplementedError

    @abstractmethod
    def get_future_notifications(self, page: int, page_size: int) -> Iterable["Notification"]:
        raise NotImplementedError

    @abstractmethod
    def get_all_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID
    ) -> Iterable["Notification"]:
        raise NotImplementedError

    @abstractmethod
    def get_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID, page: int, page_size: int
    ) -> Iterable["Notification"]:
        raise NotImplementedError

    @abstractmethod
    def persist_notification(
        self,
        user_id: int | str | uuid.UUID,
        notification_type: str,
        title: str,
        body_template: str,
        context_name: str,
        context_kwargs: dict[str, uuid.UUID | str | int],
        send_after: datetime.datetime | None,
        subject_template: str,
        preheader_template: str,
        adapter_extra_parameters: dict | None = None,
    ) -> "Notification":
        raise NotImplementedError

    @abstractmethod
    def persist_notification_update(
        self,
        notification_id: int | str | uuid.UUID,
        update_data: "UpdateNotificationKwargs",
    ) -> "Notification":
        """
        Update a notification in the backend. This method should return the updated notification.
        Notifications that have already been sent should not be updated. If a notification has already been sent,
        the method should raise a NotificationUpdateError.
        """
        raise NotImplementedError

    @abstractmethod
    def mark_pending_as_sent(self, notification_id: int | str | uuid.UUID) -> "Notification":
        raise NotImplementedError

    @abstractmethod
    def mark_pending_as_failed(self, notification_id: int | str | uuid.UUID) -> "Notification":
        raise NotImplementedError

    @abstractmethod
    def mark_sent_as_read(self, notification_id: int | str | uuid.UUID) -> "Notification":
        raise NotImplementedError

    @abstractmethod
    def cancel_notification(self, notification_id: int | str | uuid.UUID) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_notification(
        self, notification_id: int | str | uuid.UUID, for_update=False
    ) -> "Notification":
        raise NotImplementedError

    @abstractmethod
    def filter_all_in_app_unread_notifications(
        self, user_id: int | str | uuid.UUID
    ) -> Iterable["Notification"]:
        raise NotImplementedError

    @abstractmethod
    def filter_in_app_unread_notifications(
        self, user_id: int | str | uuid.UUID, page: int, page_size: int
    ) -> Iterable["Notification"]:
        raise NotImplementedError

    @abstractmethod
    def get_user_email_from_notification(self, notification_id: int | str | uuid.UUID) -> str:
        raise NotImplementedError

    @abstractmethod
    def store_context_used(
        self, 
        notification_id: int | str | uuid.UUID, 
        context: dict,
        adapter_import_str: str,
    ) -> None:
        raise NotImplementedError



import datetime
import uuid
from dataclasses import dataclass
from typing import TypedDict


class NotificationContextDict(dict):
    """
    A dictionary that only accepts string keys and values of types: int, float, str,
    list[NotificationContextDict], and dict[str, NotificationContextDict].
    """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.update(*args, **kwargs)

    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            self[k] = v

    def __setitem__(
        self,
        key: str,
        value: int
        | float
        | str
        | list["NotificationContextDict"]
        | dict[str, "NotificationContextDict"],
    ):
        if not isinstance(key, str):
            raise TypeError("Keys must be strings")
        if not isinstance(
            value,
            (int | float | str | list | dict),
        ):
            raise TypeError("Value must be an int, float, str, list, or dict")
        if isinstance(value, list):
            value = [self._validate_list_item(item) for item in value]
        if isinstance(value, dict):
            value = {k: self._validate_dict_value(v) for k, v in value.items()}
        super().__setitem__(key, value)

    def _validate_list_item(self, item):
        if not isinstance(item, NotificationContextDict):
            raise TypeError("List items must be SerializableDict instances")
        return item

    def _validate_dict_value(self, value):
        if not isinstance(value, NotificationContextDict):
            raise TypeError("Dict values must be SerializableDict instances")
        return value

    def copy(self) -> "NotificationContextDict":
        return self.__class__(super().copy())


@dataclass
class Notification:
    id: int | str | uuid.UUID  # noqa: A003
    user_id: int | str | uuid.UUID
    notification_type: str
    title: str
    body_template: str
    context_name: str
    context_kwargs: dict[str, int | str | uuid.UUID]
    send_after: datetime.datetime | None
    subject_template: str
    preheader_template: str
    status: str
    context_used: dict | None = None
    adapter_used: str | None = None
    adapter_extra_parameters: dict | None = None


class UpdateNotificationKwargs(TypedDict, total=False):
    title: str
    body_template: str
    context_name: str
    context_kwargs: dict[str, int | str | uuid.UUID]
    send_after: datetime.datetime | None
    subject_template: str | None
    preheader_template: str | None
    adapter_extra_parameters: dict | None



from abc import abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING

from vintasend.services.dataclasses import Notification
from vintasend.services.notification_template_renderers.base import (
    BaseNotificationTemplateRenderer,
    NotificationSendInput,
)


if TYPE_CHECKING:
    from vintasend.services.notification_service import NotificationContextDict


@dataclass
class TemplatedEmail(NotificationSendInput):
    subject: str
    body: str


class BaseTemplatedEmailRenderer(BaseNotificationTemplateRenderer):
    @abstractmethod
    def render(
        self, notification: Notification, context: "NotificationContextDict", **kwargs
    ) -> TemplatedEmail:
        raise NotImplementedError



import datetime
import uuid
from typing import Generic, TypeVar

from vintasend.constants import NotificationTypes
from vintasend.services.dataclasses import Notification, NotificationContextDict
from vintasend.services.notification_adapters.async_base import (
    AsyncBaseNotificationAdapter,
    NotificationDict,
)
from vintasend.services.notification_adapters.asyncio_base import AsyncIOBaseNotificationAdapter
from vintasend.services.notification_adapters.base import BaseNotificationAdapter
from vintasend.services.notification_backends.asyncio_base import AsyncIOBaseNotificationBackend
from vintasend.services.notification_backends.base import BaseNotificationBackend
from vintasend.services.notification_template_renderers.base_templated_email_renderer import (
    BaseTemplatedEmailRenderer,
)


B = TypeVar("B", bound=BaseNotificationBackend)
T = TypeVar("T", bound=BaseTemplatedEmailRenderer)


class FakeEmailAdapter(Generic[B, T], BaseNotificationAdapter[B, T]):
    notification_type = NotificationTypes.EMAIL
    backend: B
    template_renderer: T
    sent_emails: list[tuple["Notification", "NotificationContextDict"]] = []

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.sent_emails = []

    def send(self, notification: "Notification", context: "NotificationContextDict") -> None:
        self.template_renderer.render(notification, context)
        self.sent_emails.append((notification, context))


BAIO = TypeVar("BAIO", bound=AsyncIOBaseNotificationBackend)


class FakeAsyncIOEmailAdapter(Generic[BAIO, T], AsyncIOBaseNotificationAdapter[BAIO, T]):
    notification_type = NotificationTypes.EMAIL
    backend: BAIO
    template_renderer: T
    sent_emails: list[tuple["Notification", "NotificationContextDict"]] = []

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.sent_emails = []

    async def send(self, notification: "Notification", context: "NotificationContextDict") -> None:
        self.template_renderer.render(notification, context)
        self.sent_emails.append((notification, context))


class FakeAsyncEmailAdapter(AsyncBaseNotificationAdapter, Generic[B, T], FakeEmailAdapter[B, T]):
    notification_type = NotificationTypes.EMAIL

    def send(self, notification: "Notification", context: "NotificationContextDict") -> None:
        pass

    def delayed_send(self, notification_dict: NotificationDict, context_dict: dict) -> None:
        notification = self.notification_from_dict(notification_dict)
        context = NotificationContextDict(**context_dict)
        super().send(notification, context)

    def _convert_to_uuid(self, value: str) -> uuid.UUID | str:
        try:
            return uuid.UUID(value)
        except ValueError:
            return value

    def notification_from_dict(self, notification_dict: NotificationDict) -> "Notification":
        send_after = (
            datetime.datetime.fromisoformat(notification_dict["send_after"])
            if notification_dict["send_after"]
            else None
        )
        return Notification(
            id=(
                self._convert_to_uuid(notification_dict["id"])
                if isinstance(notification_dict["id"], str)
                else notification_dict["id"]
            ),
            user_id=(
                self._convert_to_uuid(notification_dict["user_id"])
                if isinstance(notification_dict["user_id"], str)
                else notification_dict["user_id"]
            ),
            context_kwargs={
                key: self._convert_to_uuid(value) if isinstance(value, str) else value
                for key, value in notification_dict["context_kwargs"].items()
            },
            send_after=send_after,
            status=notification_dict["status"],
            context_used=notification_dict["context_used"],
            notification_type=notification_dict["notification_type"],
            title=notification_dict["title"],
            body_template=notification_dict["body_template"],
            context_name=notification_dict["context_name"],
            subject_template=notification_dict["subject_template"],
            preheader_template=notification_dict["preheader_template"],
        )


class InvalidAdapter:
    def __init__(self, *_args, **_kwargs):
        pass



from collections.abc import Iterable

from vintasend.services.helpers import get_notification_adapters
from vintasend.services.notification_adapters.async_base import AsyncBaseNotificationAdapter
from vintasend.services.notification_service import NotificationService


def periodic_send_pending_notifications(
    notification_adapters: Iterable[tuple[str, str]] | None = None,
    backend_import_str: str | None = None,
    backend_kwargs: dict | None = None,
    config: dict | None = None,
):
    adapter_intances = get_notification_adapters(notification_adapters, backend_import_str, backend_kwargs)
    if isinstance(adapter_intances[0], AsyncBaseNotificationAdapter):
        desserialized_backend_kwargs = (
            adapter_intances[0].restore_backend_kwargs(backend_kwargs) if backend_kwargs else None
        )
        desserialized_config = adapter_intances[0].restore_config(config) if config else None
    
    NotificationService(
        notification_adapters=notification_adapters,
        notification_backend=backend_import_str,
        notification_backend_kwargs=desserialized_backend_kwargs,
        config=desserialized_config
    ).send_pending_notifications()
